@@ -34,6 +34,7 @@
 #include "CKeyState.h"
 #include <cstring>
 #include <cstdlib>
+#include "CScreen.h"
 
 //
 // CServer
@@ -47,7 +48,7 @@ CEvent::Type			CServer::s_switchInDirection  = CEvent::kUnknown;
 CEvent::Type			CServer::s_keyboardBroadcast  = CEvent::kUnknown;
 CEvent::Type			CServer::s_lockCursorToScreen = CEvent::kUnknown;
 
-CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
+CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient, CScreen* screen) :
 	m_primaryClient(primaryClient),
 	m_active(primaryClient),
 	m_seqNum(0),
@@ -71,11 +72,13 @@ CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
 	m_switchNeedsAlt(false),
 	m_relativeMoves(false),
 	m_keyboardBroadcasting(false),
-	m_lockedToScreen(false)
+	m_lockedToScreen(false),
+	m_screen(screen)
 {
 	// must have a primary client and it must have a canonical name
 	assert(m_primaryClient != NULL);
 	assert(config.isScreen(primaryClient->getName()));
+	assert(m_screen != NULL);
 
 	CString primaryName = getName(primaryClient);
 
@@ -127,6 +130,22 @@ CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
 							m_primaryClient->getEventTarget(),
 							new TMethodEventJob<CServer>(this,
 								&CServer::handleWheelEvent));
+	EVENTQUEUE->adoptHandler(IPlatformScreen::getGameDeviceButtonsEvent(),
+							m_primaryClient->getEventTarget(),
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleGameDeviceButtons));
+	EVENTQUEUE->adoptHandler(IPlatformScreen::getGameDeviceSticksEvent(),
+							m_primaryClient->getEventTarget(),
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleGameDeviceSticks));
+	EVENTQUEUE->adoptHandler(IPlatformScreen::getGameDeviceTriggersEvent(),
+							m_primaryClient->getEventTarget(),
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleGameDeviceTriggers));
+	EVENTQUEUE->adoptHandler(IPlatformScreen::getGameDeviceTimingReqEvent(),
+							m_primaryClient->getEventTarget(),
+							new TMethodEventJob<CServer>(this,
+								&CServer::handleGameDeviceTimingReq));
 	EVENTQUEUE->adoptHandler(IPlatformScreen::getScreensaverActivatedEvent(),
 							m_primaryClient->getEventTarget(),
 							new TMethodEventJob<CServer>(this,
@@ -308,7 +327,7 @@ CServer::adoptClient(CBaseClientProxy* client)
 
 	// send notification
 	CServer::CScreenConnectedInfo* info =
-		CServer::CScreenConnectedInfo::alloc(getName(client));
+		new CServer::CScreenConnectedInfo(getName(client));
 	EVENTQUEUE->addEvent(CEvent(CServer::getConnectedEvent(),
 								m_primaryClient->getEventTarget(), info));
 }
@@ -326,6 +345,18 @@ CServer::disconnect()
 	}
 }
 
+void
+CServer::gameDeviceTimingResp(UInt16 freq)
+{
+	m_screen->gameDeviceTimingResp(freq);
+}
+
+void
+CServer::gameDeviceFeedback(GameDeviceID id, UInt16 m1, UInt16 m2)
+{
+	m_screen->gameDeviceFeedback(id, m1, m2);
+}
+
 UInt32
 CServer::getNumClients() const
 {
@@ -1338,6 +1369,36 @@ CServer::handleWheelEvent(const CEvent& event, void*)
 	onMouseWheel(info->m_xDelta, info->m_yDelta);
 }
 
+void
+CServer::handleGameDeviceButtons(const CEvent& event, void*)
+{
+	IPlatformScreen::CGameDeviceButtonInfo* info =
+		reinterpret_cast<IPlatformScreen::CGameDeviceButtonInfo*>(event.getData());
+	onGameDeviceButtons(info->m_id, info->m_buttons);
+}
+
+void
+CServer::handleGameDeviceSticks(const CEvent& event, void*)
+{
+	IPlatformScreen::CGameDeviceStickInfo* info =
+		reinterpret_cast<IPlatformScreen::CGameDeviceStickInfo*>(event.getData());
+	onGameDeviceSticks(info->m_id, info->m_x1, info->m_y1, info->m_x2, info->m_y2);
+}
+
+void
+CServer::handleGameDeviceTriggers(const CEvent& event, void*)
+{
+	IPlatformScreen::CGameDeviceTriggerInfo* info =
+		reinterpret_cast<IPlatformScreen::CGameDeviceTriggerInfo*>(event.getData());
+	onGameDeviceTriggers(info->m_id, info->m_t1, info->m_t2);
+}
+
+void
+CServer::handleGameDeviceTimingReq(const CEvent& event, void*)
+{
+	onGameDeviceTimingReq();
+}
+
 void
 CServer::handleScreensaverActivatedEvent(const CEvent&, void*)
 {
@@ -1678,7 +1739,7 @@ CServer::onMouseUp(ButtonID id)
 bool
 CServer::onMouseMovePrimary(SInt32 x, SInt32 y)
 {
-	LOG((CLOG_DEBUG2 "onMouseMovePrimary %d,%d", x, y));
+	LOG((CLOG_DEBUG4 "onMouseMovePrimary %d,%d", x, y));
 
 	// mouse move on primary (server's) screen
 	if (m_active != m_primaryClient) {
@@ -1927,6 +1988,34 @@ CServer::onMouseWheel(SInt32 xDelta, SInt32 yDelta)
 	m_active->mouseWheel(xDelta, yDelta);
 }
 
+void
+CServer::onGameDeviceButtons(GameDeviceID id, GameDeviceButton buttons)
+{
+	LOG((CLOG_DEBUG1 "onGameDeviceButtons id=%d buttons=%d", id, buttons));
+	m_active->gameDeviceButtons(id, buttons);
+}
+
+void
+CServer::onGameDeviceSticks(GameDeviceID id, SInt16 x1, SInt16 y1, SInt16 x2, SInt16 y2)
+{
+	LOG((CLOG_DEBUG1 "onGameDeviceSticks id=%d s1=%+d,%+d s2=%+d,%+d", id, x1, y1, x2, y2));
+	m_active->gameDeviceSticks(id, x1, y1, x2, y2);
+}
+
+void
+CServer::onGameDeviceTriggers(GameDeviceID id, UInt8 t1, UInt8 t2)
+{
+	LOG((CLOG_DEBUG1 "onGameDeviceTriggers id=%d t1=%d t2=%d", id, t1, t2));
+	m_active->gameDeviceTriggers(id, t1, t2);
+}
+
+void
+CServer::onGameDeviceTimingReq()
+{
+	LOG((CLOG_DEBUG1 "onGameDeviceTimingReq"));
+	m_active->gameDeviceTimingReq();
+}
+
 bool
 CServer::addClient(CBaseClientProxy* client)
 {
@@ -2173,22 +2262,6 @@ CServer::CSwitchInDirectionInfo::alloc(EDirection direction)
 	return info;
 }
 
-
-//
-// CServer::CScreenConnectedInfo
-//
-
-CServer::CScreenConnectedInfo*
-CServer::CScreenConnectedInfo::alloc(const CString& screen)
-{
-	CScreenConnectedInfo* info =
-		(CScreenConnectedInfo*)malloc(sizeof(CScreenConnectedInfo) +
-								screen.size());
-	strcpy(info->m_screen, screen.c_str());
-	return info;
-}
-
-
 //
 // CServer::CKeyboardBroadcastInfo
 //
