@@ -18,20 +18,21 @@
 #include <gtest/gtest.h>
 #include "CArch.h"
 #include "CLog.h"
-#include "LogOutputters.h"
+
+#if SYSAPI_WIN32
+#include "CArchMiscWindows.h"
+#endif
 
 int
 main(int argc, char **argv)
 {
-	CArch arch;
-
 #if SYSAPI_WIN32
-	// only add std output logger for windows (unix
-	// already outputs to standard streams).
-	CStdLogOutputter stdLogOutputter;
-	CLOG->insert(&stdLogOutputter, true);
+	// HACK: shouldn't be needed, but logging fails without this.
+	CArchMiscWindows::setInstanceWin32(GetModuleHandle(NULL));
 #endif
 
+	CArch arch;
+
 	CLOG->setFilter(kDEBUG2);
 
 	testing::InitGoogleTest(&argc, argv);
